import logging as log

import statsd


class Base:
    """Base class for other "minty" classes

    This base class provides lazy-loaded "self.logger" and "self.statsd"
    properties.
    """

    __slots__ = ["_logger", "_statsd"]

    @property
    def logger(self):
        """Return this object's logger instance, create one if necessary

        :return: A logger object for this instance
        :rtype: logging.Logger
        """
        try:
            self._logger
        except AttributeError:
            self._logger = log.getLogger(self.__class__.__name__)

        return self._logger

    @property
    def statsd(self):
        """Return this object's statsd instance, create one if necessary

        :return: A statsd object for this instance
        :rtype: statsd.Client
        """
        try:
            self._statsd
        except AttributeError:
            class_name = self.__class__.__name__
            if class_name.startswith("_"):
                class_name = class_name[1:]
            self._statsd = statsd.Client(class_name)

        return self._statsd



__version__ = "0.0.1"

from minty import Base
from minty.exceptions import ConfigurationConflict

from .swiftwrapper import SwiftWrapper

DEFAULT_CHUNK_SIZE = 1024 * 1024 * 1024  # 1 gigabyte


class SwiftInfrastructure(Base):
    """Infrastructure Class for OpenStack/Swift Connection."""

    def __call__(self, config):
        """Create a new Swift connection using the specified configuration

        :param config: The configuration params necessary to connect to a swift container.
        :return: A Swift handle for a bucket on a connection to an OpenStack server.
        :rtype: SwiftWrapper
        """

        try:
            container_name = config["storage_bucket"]
        except KeyError:
            try:
                container_name = config["instance_uuid"]
            except KeyError as k:
                raise ConfigurationConflict(
                    "No container name specified for Swift configuration"
                ) from k

        try:
            filestore_config = config["filestore"]
            if type(filestore_config) is not list:
                filestore_config = [filestore_config]
        except KeyError as error:
            raise ConfigurationConflict(
                "No file store configuration specified for openstack/swift"
            ) from error

        return SwiftWrapper(
            filestore_config=filestore_config,
            container_name=container_name,
            segment_size=int(config.get("chunk_size", DEFAULT_CHUNK_SIZE)),
        )



import os
import xml.etree.ElementTree as ET
import zipfile

import magic
from keystoneauth1 import session
from keystoneauth1.identity import v3
from swiftclient import Connection

from minty import Base
from minty.exceptions import ConfigurationConflict


def is_zip(file_handle):
    header = file_handle.read(4)
    file_handle.seek(0, os.SEEK_SET)

    if header == b"PK\x03\x04":
        return True
    return False


def get_mime_type_from_content(file_handle):
    file_handle.seek(0, os.SEEK_SET)

    # This is needed to resolve an issue with the file handler
    # pointer not being reset despite the seek call
    file_content = file_handle.read(1024 * 1024)
    mime_type = magic.detect_from_content(file_content).mime_type

    return mime_type


def get_mime_type_from_handle(file_handle):
    recognized_office_formats = {
        "application/vnd.openxmlformats-officedocument.wordprocessingml.document.main+xml": "application/vnd.openxmlformats-officedocument.wordprocessingml.document",
        "application/vnd.openxmlformats-officedocument.spreadsheetml.sheet.main+xml": "application/vnd.openxmlformats-officedocument.spreadsheetml.sheet",
    }

    mime_type = "application/octet-stream"
    try:
        if is_zip(file_handle):
            zip_handle = zipfile.ZipFile(file_handle)
            xml_content = zip_handle.read("[Content_Types].xml")
            namespaces = {
                "ct": "http://schemas.openxmlformats.org/package/2006/content-types"
            }
            root = ET.fromstring(xml_content)
            content_types = [
                itm.attrib["ContentType"]
                for itm in root.findall("ct:Override", namespaces)
            ]

            for content_type in content_types:
                if content_type in recognized_office_formats:
                    mime_type = recognized_office_formats[content_type]
                    break
        else:
            file_handle.seek(0, os.SEEK_SET)
            mime_type = get_mime_type_from_content(file_handle)
    except (KeyError, zipfile.BadZipFile, ET.ParseError):
        file_handle.seek(0, os.SEEK_SET)
        mime_type = get_mime_type_from_content(file_handle)

    file_handle.seek(0, os.SEEK_SET)
    return mime_type


class SwiftWrapper(Base):
    """A SwiftWrapper for file actions(upload/download/remove) in Swift"""

    def __init__(
        self, filestore_config: list, container_name: str, segment_size: int
    ):
        """A swift wrapper handle to upload files to openstack-swift server in chunks.
        Returns a dictionary of the uploaded file data:  uuid, md5, file_size, mime_type, storage_location.

        Constructor for swift wrapper. It needs a connection, container name and segment_size.
        Default segments size `DEFAULT_CHUNK_SIZE` in SwiftInfrastructure.

        :param filestore_config: File store config params used to make the connection for upload.
        :type filestore_config: list
        :param container_name: The name of the container to connect to
        :type container_name: str
        :param segment_size: The size of the segment/chunk size of the file to send to swift when uploading
        :type segment_size: int
        """

        self.filestore_config = filestore_config
        self.segment_size = segment_size
        self.container_name = container_name

    def upload(self, file_handle, uuid):
        """Upload file to Swift in `segment_size` chunks.

        Reads the file handle, and uploads chunks to swift, while also
        calculating the MD5 of the file and determining its MIME type.

        :param file_handle: File-like object
        :param uuid: The uuid of the file te be handled
        :type uuid: uuid4
        :return: A dictionary of the data/metadata of the uploaded file into
            swift. Contains the keys: uuid, md5, file_size, mime_type, storage_location.
        :rtype: dict
        """
        timer = self.statsd.get_timer("file_upload")
        try:
            with timer.time("connect_to_swift"):
                connection = self._connect_to_swift(self.filestore_config[0])
        except IndexError as error:
            raise ConfigurationConflict(
                "No config found for Swift configuration"
            ) from error

        try:
            storage_location = self.filestore_config[0]["name"]
        except KeyError as error:
            raise ConfigurationConflict(
                "No name found for Swift configuration"
            ) from error

        response_dict = {}
        # For files that are smaller than 512k waitress will
        # send a BytesIO instead of a BufferedRandom handler

        mime_type = get_mime_type_from_handle(file_handle)

        with timer.time("total"):
            connection.put_object(
                container=self.container_name,
                obj=str(uuid),
                contents=file_handle,
                response_dict=response_dict,
            )

        file_handle.seek(0, os.SEEK_END)
        total_size = file_handle.tell()

        return {
            "uuid": uuid,
            "md5": response_dict["headers"]["etag"],
            "size": total_size,
            "mime_type": mime_type,
            "storage_location": storage_location,
        }

    def _connect_to_swift(self, file_store_config: dict):
        """Returns a connection to Swift based on the auth version.
        Supports V3, V2 and V1(Legacy)

        :param file_store_config: Configuration params for a file storage.
        :type file_store_config: dict
        :return: Connection V1 or V2 or V3 instance depending on the config params provided.
        :rtype: Connection
        """
        if file_store_config.get("auth") is None:
            raise ConfigurationConflict("Authentication params are missing")

        if file_store_config.get("auth_version") is None:
            raise ConfigurationConflict(
                "No auth_version specified for Swift configuration"
            )

        auth_config = {**file_store_config["auth"]}
        auth_config["timeout"] = int(auth_config.get("timeout", 60))
        auth_version = file_store_config["auth_version"]
        if auth_version == "v3":
            return self._connect_to_swift_v3(**auth_config, auth_version="3")
        elif auth_version in ["v1", "v2"]:
            # authentication version 2 and 1
            # accessing second index of string auth_version to get the version number
            auth_version = auth_version[1]
            return self._connect_to_swift_legacy_auth(
                **auth_config, auth_version=auth_version
            )
        else:
            raise ConfigurationConflict(
                f"Unsupported auth_version: '{auth_version}'"
            )

    def _connect_to_swift_v3(
        self,
        auth_url,
        username,
        password,
        auth_version=None,
        user_domain_name=None,
        project_name=None,
        project_domain_name=None,
        timeout=60,
    ) -> Connection:
        """A V3 auth type with keystone session to connect to swift.

        :param auth_url:
        :param username:
        :param password:
        :param user_domain_name:
        :param project_name:
        :param project_domain_name:
        :param timeout:
        :return  A connection Instance of V3 swift auth.
        :rtype: Connection
        """
        auth = v3.Password(
            auth_url=auth_url,
            username=username,
            password=password,
            user_domain_name=user_domain_name,
            project_name=project_name,
            project_domain_name=project_domain_name,
        )

        keystone_session = session.Session(auth=auth)

        return Connection(
            session=keystone_session,
            auth_version=auth_version,
            timeout=timeout,
        )

    def _connect_to_swift_legacy_auth(
        self,
        auth_url,
        username,
        password,
        auth_version=None,
        tenant_name=None,
        timeout=60,
    ) -> Connection:
        """Support for  auth type V1 & V2 to connect to swift.

        :param auth_url:
        :param username:
        :param password:
        :param auth_version:
        :param tenant_name:
        :param timeout:
        :return: A connection Instance of Legacy(V1) or V2 swift auth.
        :rtype: Connection
        """

        return Connection(
            authurl=auth_url,
            user=username,
            key=password,
            auth_version=auth_version,
            tenant_name=tenant_name,
            timeout=timeout,
        )



import threading
from contextlib import contextmanager

_mdc = threading.local()


def set_mdc(key, value):
    """Set a value in the MDC

    The MDC's contents are added to each log message.

    :type key: str
    :type value: object
    """
    setattr(_mdc, key, value)


def get_mdc():
    """Get an item from the MDC

    :rtype: dict from str to object
    """
    results = {}
    for log_field in dir(_mdc):
        if log_field.startswith("__"):
            continue
        results[log_field] = getattr(_mdc, log_field)
    return results


def clear_mdc():
    """Clear the MDC"""
    for k in get_mdc():
        delattr(_mdc, k)


@contextmanager
def mdc(**kw_args):
    """Context manager that adds the specified keys/values to the MDC, and
    clears them afterwards"""
    for k in kw_args:
        v = kw_args[k]
        set_mdc(k, v)

    yield

    clear_mdc()



from abc import ABC, abstractmethod
from typing import List


class EntityBase(ABC):
    @property
    @abstractmethod
    def entity_id(self):
        raise NotImplementedError

    @property
    def event_service(self):
        return self._event_service

    @event_service.setter
    def event_service(self, event_service):
        super().__setattr__("_event_service", event_service)

    @property
    def change_log(self):
        try:
            self._changes
        except AttributeError:
            self.clear_change_log()
        return self._changes

    def clear_change_log(self):
        super().__setattr__("_changes", [])

    @property
    def entity_data(self):
        try:
            self._entity_data
        except AttributeError:
            self.clear_entity_data()
        return self._entity_data

    def clear_entity_data(self):
        super().__setattr__("_entity_data", {})

    def __setattr__(self, attr, value):
        try:
            old_value = self.__getattribute__(attr)
            change = {
                "key": attr,
                "old_value": _reflect(old_value),
                "new_value": _reflect(value),
            }

            self.change_log.append(change)
        except AttributeError:
            pass
        super().__setattr__(attr, value)

    def capture_field_values(self, fields: List):
        for field in fields:
            value = self.__getattribute__(field)
            self.entity_data[field] = _reflect(value)


def _reflect(value):
    """Reflect on attribute type and return JSON parse-able type.

    :param attr: attribute
    :return: converted object to correct type
    :rtype: str, int or None
    """
    if value is None:
        return value
    elif isinstance(value, bool):
        return value
    elif isinstance(value, int):
        return int(value)
    elif isinstance(value, EntityBase):
        return {
            "type": value.__class__.__name__,
            "entity_id": str(value.entity_id),
        }
    elif isinstance(value, List):
        return [_reflect(i) for i in value]
    else:
        return str(value)



from . import Base
from .infrastructure import InfrastructureFactory


class RepositoryFactory(Base):
    """Create context-specific "repository" instances for domains
    """

    __slots__ = ["infrastructure_factory", "repositories"]

    def __init__(self, infra_factory: InfrastructureFactory):
        """Initialize the repository factory with an infrastructure factory

        :param infra_factory: Infrastructure factory
        :type infra_factory: InfrastructureFactory
        """
        self.infrastructure_factory = infra_factory
        self.repositories = {}

    def register_repository(self, name: str, repository: object):
        """Register a repository class with the repository factory.

        :param repository: repository class; will be instantiated when the
            domain code asks for it by name.
        :type repository: object
        """
        self.repositories[name] = repository

    def get_repository(self, name: str, context=None, event_service=None):
        """Retrieve a repository, given a name and optionally a context.

        :param name: name of repository to instantiate
        :type repository: str
        :param context: Context for which to retrieve the repository.
        :type context: object, optional
        :param event_service: Event service instance
        :type event_service: EventService
        :return: An instance of the configured repository, for the specified
            context.
        :rtype: object
        """
        repo_class = self.repositories[name]

        self.logger.debug(
            f"Creating repository of type '{name}' with context "
            + f"'{context}'"
        )

        with self.statsd.get_timer("get_repository").time(name):
            repo = repo_class(
                context=context,
                infrastructure_factory=self.infrastructure_factory,
                event_service=event_service,
            )

        return repo


class RepositoryBase(Base):
    """Base class for repositories using the Minty CQRS/DDD framework"""

    __slots__ = ["infrastructure_factory", "context", "cache", "event_service"]

    def __init__(
        self,
        infrastructure_factory: InfrastructureFactory,
        context: str,
        event_service,
    ):
        self.infrastructure_factory = infrastructure_factory
        self.context = context
        self.cache = {}
        self.event_service = event_service

    def _get_infrastructure(self, name):
        return self.infrastructure_factory.get_infrastructure(
            context=self.context, infrastructure_name=name
        )



import os

from .base import ConfigStoreBase


def _config_files(path):
    for dir_entry in os.scandir(path):
        if dir_entry.is_file() and dir_entry.name.endswith(".conf"):
            yield dir_entry.path


class FileStore(ConfigStoreBase):
    __slots__ = ["base_directory", "configuration"]

    def __init__(self, parser, arguments):
        """Initialize connection to filestore

        :param parser: Configuration file/data parser
        :type parser: minty.config.parser.Base
        :param directory: location of configuration files
        :type directory: str
        """

        self.parser = parser
        self.base_directory = arguments["directory"]

        self._load_config()

    def _load_config(self):
        new_config = {}

        for configfile in _config_files(self.base_directory):
            self.logger.info(f"Reading configuration from '{configfile}")
            with open(configfile, "r", encoding="utf-8") as file:
                content = file.read()

            config_data = self.parser.parse(content)

            for (key, instance_config) in config_data.items():
                self.logger.info(f"Found main configuration '{key}'")
                new_config[key] = instance_config

                try:
                    aliases = [
                        alias.strip()
                        for alias in instance_config["aliases"].split(",")
                    ]
                except KeyError:
                    # "aliases" is an optional configuration item
                    aliases = []

                for alias in aliases:
                    self.logger.info(f"Found alias configuration '{alias}'")
                    new_config[alias] = instance_config

        self.configuration = new_config

    def retrieve(self, name: str) -> str:
        """Retrieve configuration from a file

        :param name: base name of the configuration file
        :type name: str
        :return: configuration file contents
        :rtype: str
        """

        super().retrieve(name)

        return self.configuration[name]



from abc import ABC, abstractmethod
from functools import partial
from uuid import UUID

from .. import Base
from ..infrastructure import InfrastructureFactory
from ..repository import RepositoryFactory
from .events import Event, EventService, event

__all__ = ["event", "Event", "EventService"]


class CommandBase(Base, ABC):
    __slots__ = ["repository_factory", "context", "user_uuid"]

    def __init__(
        self,
        repository_factory: RepositoryFactory,
        context: str,
        user_uuid: UUID,
        event_service: EventService,
    ):
        self.repository_factory = repository_factory
        self.context = context
        self.user_uuid = user_uuid
        self.event_service = event_service

    def get_repository(self, name):
        """Initialize and retrieve a repository by name.

        This also passes the event service to the repository constructor.

        :param name: Name of the repository to retrieve
        :type name: str
        :return: Initialized repository
        :rtype: RepositoryBase
        """

        return self.repository_factory.get_repository(
            name=name, context=self.context, event_service=self.event_service
        )


class QueryBase(Base, ABC):
    __slots__ = ["repository_factory", "context", "user_uuid"]

    def __init__(self, repository_factory, context, user_uuid: UUID):
        self.repository_factory = repository_factory
        self.context = context
        self.user_uuid = user_uuid

    def get_repository(self, name):
        """Initialize and retrieve a repository by name.

        Passes None as the `event_service`, so any calls to state-changing
        commands can raise the right exceptions.

        :param name: Name of the repository to retrieve
        :type name: str
        :return: Initialized repository
        :rtype: RepositoryBase
        """

        return self.repository_factory.get_repository(
            name=name, context=self.context, event_service=None
        )


class QueryMiddleware(Base, ABC):
    __slots__ = [
        "infrastructure_factory",
        "correlation_id",
        "domain",
        "context",
        "user_uuid",
    ]

    def __init__(
        self,
        infrastructure_factory: InfrastructureFactory,
        correlation_id: UUID,
        domain: str,
        context: str,
        user_uuid: UUID,
    ):
        """Initialize the middleware with an infrastructure factory and some
        parameters from the current request.

        :param infrastructure_factory: [description]
        :type infrastructure_factory: InfrastructureFactory
        :param correlation_id: Unique identifier for the current "chain of
            events"
        :type correlation_id: UUID
        :param domain: Domain the command is a part of
        :type domain: str
        :param context: Context the command is being executed in
        :type context: str
        :param user_uuid: UUID of the user executing the command
        :type user_uuid: UUID
        """

        self.infrastructure_factory = infrastructure_factory
        self.correlation_id = correlation_id
        self.domain = domain
        self.context = context
        self.user_uuid = user_uuid

    @abstractmethod
    def __call__(self, func):
        """Call the specified function.

        Override this method (including the defined params) in your middleware
        class. The logic or functionality you want to implement should go
        before and after your call to `func()`.

        Remember to return whatever `func()` returns!

        :param func: function to execute
        :type func: partial function
        """
        pass


class MiddlewareBase(Base, ABC):
    __slots__ = [
        "infrastructure_factory",
        "event_service",
        "correlation_id",
        "domain",
        "context",
        "user_uuid",
    ]

    def __init__(
        self,
        infrastructure_factory: InfrastructureFactory,
        event_service: EventService,
        correlation_id: UUID,
        domain: str,
        context: str,
        user_uuid: UUID,
    ):
        """Initialize the middleware with an infrastructure factory and some
        parameters from the current request.

        :param infrastructure_factory: Infrastructure factory to use
        :type infrastructure_factory: InfrastructureFactory
        :param event_service: Event service instance to use
        :type event_service: EventService
        :param correlation_id: Unique identifier for the current "chain of
            events"
        :type correlation_id: UUID
        :param domain: Domain the command is a part of
        :type domain: str
        :param context: Context the command is being executed in
        :type context: str
        :param user_uuid: UUID of the user executing the command
        :type user_uuid: UUID
        """

        self.infrastructure_factory = infrastructure_factory
        self.event_service = event_service
        self.correlation_id = correlation_id
        self.domain = domain
        self.context = context
        self.user_uuid = user_uuid

    @abstractmethod
    def __call__(self, func):
        """Call instantiated class.

        Override this method (including the defined params) in your middleware
        class. The logic or functionality you want to implement should go
        before and after your call to `func()`.

        :param func: function to execute
        :type func: partial function
        """
        pass


class CommandInfrastructureCleanup(MiddlewareBase):
    """Set current event in infrastructure factory and clear it after the event is finished."""

    def __call__(self, func):
        try:
            func()
        finally:
            self.infrastructure_factory.flush_local_storage()


class QueryInfrastructureCleanup(QueryMiddleware):
    """Set current event in infrastructure factory and clear it after the query is finished."""

    def __call__(self, func):
        try:
            rv = func()
            return rv
        finally:
            self.infrastructure_factory.flush_local_storage()


class QueryWrapper(Base):
    """Wrapper class for query instances that applies middleware."""

    def __init__(self, query_instance, middleware: list):
        self.middleware = middleware
        self.query_instance = query_instance

    def __getattr__(self, attr):
        """Get an attribute on the wrapped class, wrapped by "event" code.

        This event code ensures the command can't return anything, and creates
        an Event instance.

        :param attr: attribute to retrieve
        :type attr: str
        :return: wrapped method
        :rtype: callable
        """
        original_attribute = self.query_instance.__getattribute__(attr)
        if callable(original_attribute):

            def wrapped(*args, **kwargs):
                """Return attribute wrapped in middlewares."""

                with self.statsd.get_timer("TotalQueryTime").time(attr):
                    wrapped_func = partial(original_attribute, *args, **kwargs)

                    for middleware in self.middleware:
                        wrapped_func = partial(middleware, func=wrapped_func)
                    return wrapped_func()

            return wrapped
        else:
            return original_attribute


class CommandWrapper(Base):
    """Wrapper class for command instances. Handles creation of Events."""

    def __init__(self, command_instance):
        self.command_instance = command_instance
        self.middlewares = []

    def register_middleware(self, middleware: MiddlewareBase):
        """Register middleware to be wrapped around command.

        From inner to outer layer, the last middleware class to get registered will be
        the outer shell and will be executed first and last.

        :param middleware: middleware to register
        :type middleware: MiddlewareBase
        """
        self.middlewares.append(middleware)

    def __getattr__(self, attr):
        """Get an attribute on the wrapped class, wrapped by "event" code.

        This event code ensures the command can't return anything, and creates
        an Event instance.

        :param attr: attribute to retrieve
        :type attr: str
        :return: wrapped method
        :rtype: callable
        """

        original_attribute = self.command_instance.__getattribute__(attr)
        if callable(original_attribute):

            def wrapped(*args, **kwargs):
                """Return attribute wrapped in middlewares."""
                with self.statsd.get_timer("TotalCommandTime").time(attr):

                    wrapped_func = partial(original_attribute, **kwargs)
                    for middleware in self.middlewares:
                        wrapped_func = partial(middleware, func=wrapped_func)

                    wrapped_func()
                return

            return wrapped
        else:
            return original_attribute


class CQRS(Base):
    """Keep commands and queries separated.

    CQRS: Command Query Responsibility Separation
    """

    __slots__ = [
        "domains",
        "infrastructure_factory",
        "command_wrapper_middleware",
        "query_middleware",
    ]

    def __init__(
        self,
        domains,
        infrastructure_factory: InfrastructureFactory,
        command_wrapper_middleware=None,
        query_middleware=None,
    ):
        """Create a new CQRS instance from a list of domains.

        :param domains: iterable returning domains. Domains are classes or
            packages with at least a "REQUIRED_REPOSITORIES" variable defining
            which repositories are necessary to use the domain.
        :type domains: object
        :param infrastructure_factory: Infrastructure factory, created with
            the required configuration, that the repositories can use to
            create infrastructure instances.
        :type infrastructure_factory: InfrastructureFactory
        :param command_wrapper_middleware: Middlewares to be wrapped around
            command.
        :type command_wrapper_middleware: list of MiddlewareBase
        """
        self.domains = {}

        if command_wrapper_middleware is None:
            command_wrapper_middleware = []

        if query_middleware is None:
            query_middleware = []

        self.command_wrapper_middleware = command_wrapper_middleware
        self.query_middleware = query_middleware
        self.infrastructure_factory = infrastructure_factory

        for domain in domains:
            repo_factory = RepositoryFactory(infrastructure_factory)

            for name, repo in domain.REQUIRED_REPOSITORIES.items():
                repo_factory.register_repository(name=name, repository=repo)

                for name, infra in repo.REQUIRED_INFRASTRUCTURE.items():
                    infrastructure_factory.register_infrastructure(
                        name=name, infrastructure=infra
                    )

            self.domains[domain.__name__] = {
                "module": domain,
                "repository_factory": repo_factory,
            }

    def get_query_instance(
        self, correlation_id: UUID, domain: str, context, user_uuid: UUID
    ):
        """Instantiate and return the "query" part of the specified domain.

        :param domain: name of the domain to get the query instance for
        :type domain: str
        :param context: context for this query instance
        :type context: str
        :param user_uuid: UUID of the user that's going to execute commands
        :type user_uuid: uuid
        """

        self.logger.debug(
            f"Creating query instance for domain '{domain}' with context "
            + f"'{context}' for user '{user_uuid}'"
        )

        with self.statsd.get_timer(domain).time("get_query_instance"):
            query_instance = self.domains[domain]["module"].get_query_instance(
                self.domains[domain]["repository_factory"],
                context=context,
                user_uuid=user_uuid,
            )

            initialized_middleware = []

            # Outermost layer: cleanup infrastructure when we're done
            for mw in [*self.query_middleware, QueryInfrastructureCleanup]:
                initialized_middleware.append(
                    mw(
                        infrastructure_factory=self.infrastructure_factory,
                        correlation_id=correlation_id,
                        domain=domain,
                        context=context,
                        user_uuid=user_uuid,
                    )
                )

            wrapped_query = QueryWrapper(
                query_instance=query_instance,
                middleware=initialized_middleware,
            )

        return wrapped_query

    def get_command_instance(
        self, correlation_id: UUID, domain: str, context, user_uuid: UUID
    ):
        """Instantiate and return the "command" instance of the specified domain.

        Command instance is instantiated with optional layers of middleware to handle
        various functions when executing a command. `InfrastructureStateManager` is
        always registered as the outermost layer.

        :param correlation_id: unique identifier for the current chain of
            events.
        :param domain: name of the domain to get the query instance for
        :type domain: str
        :param context: context for this command instance
        :type context: str
        :param user_uuid: UUID of the user that's going to execute commands
        :type user_uuid: uuid
        """
        self.logger.debug(
            f"Creating command instance for domain '{domain}' with context "
            + f"'{context}' for user '{user_uuid}'"
        )

        with self.statsd.get_timer(domain).time("get_command_instance"):
            event_service = EventService(
                correlation_id=correlation_id,
                domain=domain,
                context=context,
                user_uuid=user_uuid,
            )

            cmd_instance = self.domains[domain]["module"].get_command_instance(
                self.domains[domain]["repository_factory"],
                context=context,
                user_uuid=user_uuid,
                event_service=event_service,
            )

            cmd_wrapped = CommandWrapper(command_instance=cmd_instance)

            # Always "outer layer" to clean up infrastructure
            for middleware in [
                *self.command_wrapper_middleware,
                CommandInfrastructureCleanup,
            ]:
                initialized_middleware = middleware(
                    infrastructure_factory=self.infrastructure_factory,
                    correlation_id=correlation_id,
                    domain=domain,
                    context=context,
                    user_uuid=user_uuid,
                    event_service=event_service,
                )
                cmd_wrapped.register_middleware(initialized_middleware)

        return cmd_wrapped



import json
from functools import wraps
from uuid import UUID

import statsd
from jsonpointer import JsonPointer
from jsonschema import Draft7Validator, draft7_format_checker

from .exceptions import ValidationError

_format_checker = draft7_format_checker


@_format_checker.checks("uuid", raises=ValueError)
def is_uuid(instance):
    return UUID(hex=str(instance))


def validate_with(schema_data: bytes):
    """Ensure the decorated function is called with valid arguments

    :param schema_data: The JSONSchema to validate against.
        Should be UTF-8-encoded JSON bytes.
    :type schema_data: bytes
    :return: Decorator for validating function arguments against `schema`
    """
    schema = json.loads(schema_data.decode("utf-8"))

    def validity_wrapper(f):
        validator = Draft7Validator(
            schema=schema, format_checker=_format_checker
        )

        @wraps(f)
        def check_validity(*args, **kwargs):
            with statsd.Timer(__name__).time("check_validity"):
                errors = []
                for error in validator.iter_errors(instance=kwargs):
                    error_pointer = JsonPointer.from_parts(error.absolute_path)

                    errors.append(
                        {
                            "context": error.validator,
                            "message": error.message,
                            "cause": str(error.cause) if error.cause else None,
                            "property": error_pointer.path,
                        }
                    )

                if len(errors):
                    raise ValidationError(errors)

            return f(*args, **kwargs)

        return check_validity

    return validity_wrapper



from dataclasses import dataclass
from datetime import datetime
from typing import List
from uuid import UUID, uuid4

from .. import Base


def event(name: str, extra_fields: List = None):
    """Event decorator to capture entity changes and publish events to the event_service.

    :param name: Name of event to publish
    :type name: str
    :param extra_fields: extra fields to capture publish with event, defaults to None
    :type extra_fields: List, optional
    :return: wrapped_entity
    :rtype: func
    """

    def register_event(func):
        def wrapper(*args, **kwargs):
            func(*args, **kwargs)
            wrapped_entity = args[0]

            if extra_fields is not None:
                wrapped_entity.capture_field_values(fields=extra_fields)

            wrapped_entity.event_service.log_event(
                entity_type=wrapped_entity.__class__.__name__,
                entity_id=wrapped_entity.entity_id,
                event_name=name,
                changes=wrapped_entity.change_log,
                entity_data=wrapped_entity.entity_data,
            )

            wrapped_entity.clear_entity_data()
            wrapped_entity.clear_change_log()

        return wrapper

    return register_event


@dataclass
class Event:
    uuid: UUID
    created_date: datetime

    correlation_id: UUID
    domain: str
    context: str
    user_uuid: UUID

    entity_type: str
    entity_id: UUID

    event_name: str
    changes: dict
    entity_data: dict


class EventService(Base):
    __slots__ = [
        "event_list",
        "correlation_id",
        "domain",
        "context",
        "user_uuid",
    ]

    def __init__(
        self, correlation_id: UUID, domain: str, context: str, user_uuid: UUID
    ):
        self.correlation_id = correlation_id
        self.domain = domain
        self.context = context
        self.user_uuid = user_uuid
        self.event_list: List[Event] = []

    def log_event(
        self, entity_type, entity_id, event_name, changes, entity_data
    ):
        """Register a new event with the event serivce

        This will create a new event using the `EventFactory` configured on
        initialization and append it to the `event_list`.

        :param event_name: Name of the event that happened
        :type event_name: str
        :param parameters: Dictionary containing the event parameters (like
            entity state pre-event and post-event)
        :type parameters: dict
        """

        uuid = uuid4()
        created_date = datetime.utcnow()

        event = Event(
            uuid=uuid,
            created_date=created_date,
            correlation_id=self.correlation_id,
            domain=self.domain,
            context=self.context,
            user_uuid=self.user_uuid,
            entity_type=entity_type,
            entity_id=entity_id,
            event_name=event_name,
            changes=changes,
            entity_data=entity_data,
        )

        self.logger.info(f"Created Event: {event}")

        self.event_list.append(event)
